@@ -1,7 +1,10 @@
 /*
- * Copyright (c) James Fidell 1994, 1995, 1996.
  *
- * Permission to use, copy, modify, distribute, and sell this software
+ * $Id: Beeb.c,v 1.35 2002/01/15 15:46:43 james Exp $
+ *
+ * Copyright (C) James Fidell 1994-2002.
+ *
+ * Permission to use, copy, modify and distribute this software
  * and its documentation for any purpose is hereby granted without fee,
  * provided that the above copyright notice appear in all copies and
  * that both that copyright notice and this permission notice appear in
@@ -22,6 +25,172 @@
  *
  */
 
+/*
+ * Modification History
+ *
+ * $Log: Beeb.c,v $
+ * Revision 1.35  2002/01/15 15:46:43  james
+ * *** empty log message ***
+ *
+ * Revision 1.34  2002/01/14 23:20:44  james
+ * Prevent rolling over 31 disk catalog entries
+ *
+ * Revision 1.33  2002/01/14 22:18:51  james
+ * Added support for .inf files
+ *
+ * Revision 1.32  2000/09/07 21:59:03  james
+ * Add FASTHOST configurable.  Make bitmap displays read data properly rather
+ * than direct from the memory array when this is enabled.
+ *
+ * Revision 1.31  2000/09/06 11:41:13  james
+ * First cut at "proper" sound code
+ *
+ * Revision 1.30  2000/08/31 23:07:41  james
+ * Updated timing mechanism to run at a more realistic speed
+ *
+ * Revision 1.29  2000/08/16 17:58:26  james
+ * Update copyright message
+ *
+ * Revision 1.28  1996/11/24 22:13:25  james
+ * Timer values need to be updated before they are read or over-written
+ * in the 6522 User and System VIA code.  From a fix by David Ralph Stacey.
+ *
+ * Revision 1.27  1996/11/15 08:36:47  james
+ * ExitEmulator label is only required for base 6502.
+ *
+ * Revision 1.26  1996/10/12 15:15:01  james
+ * Compat.h added to support non-UNIXisms, in particular Win32.
+ *
+ * Revision 1.25  1996/10/10 21:46:38  james
+ * Corrected missed conversion of 100 cycles to FASTCLOCK
+ *
+ * Revision 1.24  1996/10/10 21:45:51  james
+ * Tidy up tabbing.
+ *
+ * Revision 1.23  1996/10/09 22:06:51  james
+ * Overhaul of the bitmapped screen handling code with particular respect to
+ * colour maps.
+ *
+ * Revision 1.22  1996/10/07 22:59:45  james
+ * Modified FASTCLOCK implementation to allow configuration of the number
+ * of instructions between interrupts.
+ *
+ * Revision 1.21  1996/10/07 22:44:18  james
+ * Moved IRQ handling back into the main code rather than #defined as
+ * previously -- basically, it doesn't work that way.
+ *
+ * Revision 1.20  1996/10/07 22:06:31  james
+ * Added XDFS ROM & support code from David Ralph Stacey.
+ *
+ * Revision 1.19  1996/10/01 22:30:31  james
+ * Added VoxWare sound code from James Murray <jsm@jsm-net.demon.co.uk>.
+ *
+ * Revision 1.18  1996/10/01 22:09:57  james
+ * Split keyboard handling into kEYMAP_STRICT and KEYMAP_LEGEND models.
+ *
+ * Revision 1.17  1996/10/01 00:32:58  james
+ * Created separate hardware reset code for each emulated unit and called
+ * these from the main initialisation section of the code to do all of the
+ * setup necessary.
+ *
+ * Revision 1.16  1996/09/30 23:46:50  james
+ * Correction to load the correct language ROM
+ *
+ * Revision 1.15  1996/09/30 23:39:29  james
+ * Split out option processing into Options.[ch].  Updated the help message,
+ * added support for the Model A using the -a switch (and added the
+ * MODEL_B_ONLY #define in Config.h, added the -m and -s switches to set the
+ * initial screen mode and keyboard DIP switches.
+ *
+ * Revision 1.14  1996/09/24 23:05:34  james
+ * Update copyright dates.
+ *
+ * Revision 1.13  1996/09/24 22:40:15  james
+ * Massive overhaul of instruction decoding code.  Includes :
+ *
+ *   Correct implementation of (indirect),Y instructions when overflow occurs,
+ *   allowing the removal of the RANGE_CHECK directive for those instructions.
+ *
+ *   Correct handling of address wrap-around for all zp,X and zp,Y
+ *   instructions.  This removes the need for the RANGE_CHECK define.  Removed
+ *   that, too.
+ *
+ *   Updated all disassembly instructions to give the full number of hex
+ *   digits when displaying their parameters.
+ *
+ *   Split opcodes.h to give NMOS 6502 opcodes in 6502ops.h and EFS opcodes
+ *   in EFSops.h
+ *
+ *   Add all NMOS 6502 HALT opcodes.
+ *
+ *   Add all NMOS 6502 NOP opcodes.
+ *
+ *   Coded for the undocumented NMOS 6502 NOP operations so that
+ *   they load a value from memory according to their addressing mode (but
+ *   neither store it anywhere nor set any SR flags).
+ *
+ *   Coded all other undocumented NMOS 6502 operations.
+ *
+ *   Changed the EFS dummy opcodes because of a clash with the undocumented
+ *   NMOS 6502 DCP instructions.  The new trap values are now codes that would
+ *   normally halt the CPU.
+ *
+ *   Added all the R65C02 opcodes.
+ *
+ *   Added all the R65C12 opcodes.
+ *
+ *   Correctly coded (zp,X) addressing mode where zp+X(+1) overlaps the
+ *   page boundary.
+ *
+ *   Added #defined values for the number of cycles taken by each instruction
+ *   in 6502.h
+ *
+ *   Added #defines for the original 6502 and Rockwell 65C02 and 65C12.
+ *
+ * Revision 1.12  1996/09/23 16:33:17  james
+ * Tidied up code for exiting the emulator
+ *
+ * Revision 1.11  1996/09/23 16:21:15  james
+ * Improvements to snapshot code.
+ *
+ * Revision 1.10  1996/09/22 21:00:54  james
+ * Changed IRQ-handling code to be called only when I think an IRQ may have
+ * happened.  Converted it all to #defines so that it can be compiled inline.
+ *
+ * Revision 1.9  1996/09/22 20:28:52  james
+ * Added Patchlevel.h and -V command line option.
+ *
+ * Revision 1.8  1996/09/22 20:20:37  james
+ * Corrections to DISASS code.
+ *
+ * Revision 1.7  1996/09/22 19:23:20  james
+ * Add the emulated filing system code.
+ *
+ * Revision 1.6  1996/09/21 23:07:35  james
+ * Call FatalError() rather than exit() so that screen stuff etc. can
+ * be cleaned up.
+ *
+ * Revision 1.5  1996/09/21 22:48:21  james
+ * Add instruction counting code.
+ *
+ * Revision 1.4  1996/09/21 22:39:52  james
+ * Improved handling of instruction disassembly.
+ *
+ * Revision 1.3  1996/09/21 22:13:46  james
+ * Replaced "unsigned char" representation of 1 byte with "byteval".
+ *
+ * Revision 1.2  1996/09/21 19:04:09  james
+ * Renamed Floppy.[ch] to Disk.[ch]
+ *
+ * Revision 1.1  1996/09/21 17:20:35  james
+ * Source files moved to src directory.
+ *
+ * Revision 1.1.1.1  1996/09/21 13:52:48  james
+ * Xbeeb v0.1 initial release
+ *
+ *
+ */
+
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -30,6 +199,7 @@
 #include <unistd.h>
 #include <string.h>
 #include <limits.h>
+#include <sys/time.h>
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 
@@ -58,20 +228,51 @@
 #include "VideoUla.h"
 #include "Options.h"
 #include "Sound.h"
+#include "Compat.h"
 
 #ifdef	EMUL_FS
-#include "EFS.h"
+#include "EmulFS.h"
 #endif
 
 
+/*
+ * Obviously this has to come first so that we know if a forward
+ * declaration of the AddClockCycles function is required.
+ */
+
+#ifdef  FASTCLOCK
+
+#define AddClockCycles(n) \
+{ \
+	ClockCyclesSoFar += n; \
+	Cycles += n; \
+	if ( ClockCyclesSoFar > FASTCLOCK ) \
+	{ \
+		ViaClockUpdate ( ClockCyclesSoFar ); \
+		ClockCyclesSoFar = 0; \
+	} \
+}
+
+#else   /* FASTCLOCK */
+
+#define AddClockCycles(n) \
+	ViaClockUpdate(n); \
+	Cycles += n;
+
+#endif  /* FASTCLOCK */
+
+
 /*
  * Misc. forward declarations...
  */
 
-void				SaveSnapshot();
-void				RestoreSnapshot ( char* );
-void				Beeb();
-static void			HardwareReset();
+void						SaveSnapshot();
+void						RestoreSnapshot ( char* );
+void						Beeb();
+static void					HardwareReset();
+#ifndef	AddClockCycles
+void						AddClockCycles ( byteval );
+#endif
 
 /*
  * This lot are just for convenience so that the real variables can
@@ -79,12 +280,12 @@ static void			HardwareReset();
  * registers.
  */
 
-byteval 			InitAccumulator;
-byteval				InitRegisterX;
-byteval				InitRegisterY;
-byteval				InitStackPointer;
-unsigned int		InitProgramCounter;
-int					NewPC;
+byteval						InitAccumulator;
+byteval						InitRegisterX;
+byteval						InitRegisterY;
+byteval						InitStackPointer;
+unsigned int				InitProgramCounter;
+int							NewPC;
 
 /*
  * The status register and the actual values of all the flags.  The
@@ -92,41 +293,41 @@ int					NewPC;
  * put in a register.
  */
 
-byteval				StatusRegister;
+byteval						StatusRegister;
 
-byteval				NegativeFlag = 0;
-byteval				OverflowFlag = 0;
-byteval				DecimalModeFlag = 0;
-byteval				IRQDisableFlag = 0;
-byteval				ZeroFlag = 0;
-byteval				CarryFlag = 0;
+byteval						NegativeFlag = 0;
+byteval						OverflowFlag = 0;
+byteval						DecimalModeFlag = 0;
+byteval						IRQDisableFlag = 0;
+byteval						ZeroFlag = 0;
+byteval						CarryFlag = 0;
 
 /*
  * The 3 jump addresses used by the 6502 -- set up at initialisation
  * once the OS ROM has been loaded.
  */
 
-unsigned int		NMIAddress;
-unsigned int		IRQAddress;
-unsigned int		ResetAddress;
+unsigned int				NMIAddress;
+unsigned int				IRQAddress;
+unsigned int				ResetAddress;
 
 /*
  * Names of files to load which are passed in as command line options.
  */
 
-char				*OsRomName = OS_ROM;
-char				*Rom15Name = LANG_ROM;
+char						*OsRomName = OS_ROM;
+char						*Rom15Name = LANG_ROM;
 #ifdef	XDFS
-char				*Rom14Name = XDFS_ROM;
+char						*Rom14Name = XDFS_ROM;
 #endif
-char				*SnapshotName;
+char						*SnapshotName;
 
-byteval				SnapshotRequested = 0;
-byteval				QuitEmulator = 0;
-byteval				BreakKeypress = 0;
+unsigned char				SnapshotRequested = 0;
+unsigned char				QuitEmulator = 0;
+unsigned char				BreakKeypress = 0;
+
+unsigned char				MaskableInterruptRequest = 0;
 
-byteval 			MaskableInterruptRequest = 0;
-static byteval 		ClockCyclesSoFar = 0;
 
 /*
  * Debugging stuff
@@ -142,32 +343,11 @@ unsigned char		DebugLevel = 0;
 static unsigned long	InstrCount [ 256 ];
 #endif
 
-
-#ifdef	FASTCLOCK
-
-#define	AddClockCycles(n) \
-{ \
-	ClockCyclesSoFar += n; \
-	if ( ClockCyclesSoFar > FASTCLOCK ) \
-	{ \
-		ViaClockUpdate ( ClockCyclesSoFar ); \
-		ClockCyclesSoFar = 0; \
-	} \
-}
-
-#else	/* FASTCLOCK */
-
-#define	AddClockCycles(n)	ViaClockUpdate(n)
-
-#endif	/* FASTCLOCK */
-
-
-
 int
 main ( int argc, char** argv )
 {
 #ifdef	COUNT_INSTRS
-	int			i;
+	int				i;
 #endif
 
 	ProcessOptions ( argc, argv );
@@ -222,7 +402,13 @@ Beeb()
 	register byteval		RegisterY;
 	register byteval		StackPointer;
 
+	unsigned int			Cycles;
 	unsigned int			OpCount = 0;
+#ifdef	FASTHOST
+	unsigned long			usec = 0;
+	long					timediff = 0;
+	struct timeval			tv;
+#endif
 
 	/*
 	 * Now load a snapshot if one was specified.
@@ -239,10 +425,23 @@ Beeb()
 	StackPointer = InitStackPointer;
 	SetProgramCounter ( InitProgramCounter );
 
-	while ( 1 )
+#ifdef	LIMIT
+#define	CONT_CONDITION		OpCount < LIMIT
+#else
+#define	CONT_CONDITION		1
+#endif
+
+	Cycles = 0;
+
+#if defined(FASTHOST) && !defined(FASTCLOCK)
+		( void ) gettimeofday ( &tv, 0 );
+		usec = tv.tv_usec;
+#endif
+
+	while ( CONT_CONDITION )
 	{
 
-		unsigned int	opcode;
+		byteval			opcode;
 
 /*
  * The 6502 execute instruction loop is in a different source file
@@ -251,6 +450,13 @@ Beeb()
 
 #include "6502.c"
 
+#ifdef	FASTHOST
+		( void ) gettimeofday ( &tv, 0 );
+		if ( tv.tv_sec > SoundCallSecs || ( tv.tv_sec == SoundCallSecs
+				&& tv.tv_usec >= SoundCallUsecs ))
+			SoundRefresh();
+#endif
+
 		/*
 		 * Now check to see if we need to service an interrupt
 		 */
@@ -293,12 +499,14 @@ Beeb()
 				ViaClockUpdate ( ClockCyclesSoFar );
 
 				SaveSnapshot();
-				SnapshotRequested = 0;
+                SnapshotRequested = 0;
 			}
 			else
+			{
 				if ( QuitEmulator )
 					break;
 				else
+				{
 					if ( BreakKeypress )
 					{
 						do
@@ -309,15 +517,32 @@ Beeb()
 						HardwareReset();
 						SetProgramCounter ( InitProgramCounter );
 					}
+				}
+			}
 		}
+#if defined(FASTHOST) && !defined(FASTCLOCK)
+		/*
+		 * Wait for amount of time this instruction should have taken
+		 */
 
-#ifdef	LIMIT
-	} while ( OpCount < LIMIT );
-#else
-	}
+		if ( Cycles > 1000 ) {
+			do {
+				( void ) gettimeofday ( &tv, 0 );
+				timediff = tv.tv_usec - usec;
+				if ( timediff < 0 ) timediff += 1000000;
+				if ( tv.tv_sec > SoundCallSecs || ( tv.tv_sec == SoundCallSecs
+						&& tv.tv_usec >= SoundCallUsecs ))
+					SoundRefresh();
+			} while ( timediff < ( Cycles / CPU_SPEED ));
+			usec = tv.tv_usec;
+			Cycles = 0;
+		}
 #endif
+	}
 
+#ifdef	M6502
 ExitEmulator:
+#endif
 
 	ShutdownScreen();
 	return;
@@ -370,6 +595,9 @@ void
 AddClockCycles ( byteval val )
 {
 	ClockCyclesSoFar += val;
+	Cycles += val;
+	SoundCycles -= val;
+	SoundCounter += val;
 	if ( ClockCyclesSoFar > FASTCLOCK )
 	{
 		ViaClockUpdate ( ClockCyclesSoFar );
@@ -385,6 +613,9 @@ void
 AddClockCycles ( byteval val )
 {
 	ViaClockUpdate ( val );
+	Cycles += val;
+	SoundCycles -= val;
+	SoundCounter += val;
 	return;
 }
 
@@ -393,7 +624,7 @@ AddClockCycles ( byteval val )
 
 
 void
-SaveSnapshot ()
+SaveSnapshot()
 {
 	int						fd, done, ilen, xlen;
 	static unsigned char	MagicNo[4] = { 0x0b, 0xbc, 0x00, 0x01 };
@@ -512,7 +743,8 @@ SaveSnapshot ()
 		}
 	}
 
-	if (( fd = open ( InputBuffer, O_WRONLY | O_TRUNC | O_CREAT, 0644 )) < 0 )
+	if (( fd = open ( InputBuffer, O_WRONLY | O_TRUNC | O_CREAT | O_BINARY,
+																0644 )) < 0 )
 	{
 		fprintf ( stderr, "Couldn't create snapshot file\n" );
 		FatalError();
@@ -651,7 +883,7 @@ RestoreSnapshot ( char* sname )
 	 * Try locally first, then prepend the snapshot directory.
 	 */
 
-	if (( fd = open ( SnapName, O_RDONLY )) < 0 )
+	if (( fd = open ( SnapName, O_RDONLY | O_BINARY )) < 0 )
 	{
 		if (( l + strlen ( XBEEBSNAPS )) <= PATH_MAX )
 		{
@@ -778,13 +1010,13 @@ IRQ()
 {
 	MaskableInterruptRequest++;
 }
-#endif
+#endif	/* IRQ */
 
 
 int
 SaveCPU ( int fd )
 {
-	byteval		cpu [ 16 ];
+	byteval			cpu [ 16 ];
 
 	GenerateStatusRegister;
 	cpu [ 0 ] = InitAccumulator;
@@ -806,7 +1038,7 @@ SaveCPU ( int fd )
 int
 RestoreCPU ( int fd, unsigned int ver )
 {
-	byteval		cpu [ 16 ];
+	byteval			cpu [ 16 ];
 
 	if ( ver > 1 )
 		return -1;
