@@ -1,7 +1,10 @@
 /*
- * Copyright (c) James Fidell 1994.
  *
- * Permission to use, copy, modify, distribute, and sell this software
+ * $Id: Bitmap.h,v 1.8 2002/01/15 15:46:43 james Exp $
+ *
+ * Copyright (C) James Fidell 1994-2002.
+ *
+ * Permission to use, copy, modify and distribute this software
  * and its documentation for any purpose is hereby granted without fee,
  * provided that the above copyright notice appear in all copies and
  * that both that copyright notice and this permission notice appear in
@@ -22,7 +25,31 @@
  *
  */
 
+/*
+ * Modification History
+ *
+ * $Log: Bitmap.h,v $
+ * Revision 1.8  2002/01/15 15:46:43  james
+ * *** empty log message ***
+ *
+ * Revision 1.7  2000/08/16 17:58:26  james
+ * Update copyright message
+ *
+ * Revision 1.6  1996/11/09 23:02:28  james
+ * Take account of the horizontal start of display based on the clock rate
+ * set in the Video ULA and the CRTC Horiz. Sync. register.
+ *
+ * Revision 1.5  1996/11/07 23:41:37  james
+ * Add CVS headers.
+ *
+ *
+ */
 
+
+#ifndef	BITMAP_H
+#define	BITMAP_H
+
+extern	void			InitialiseBitmap();
 extern	void			BitmapScanlineUpdate ( unsigned int );
 extern	void			BitmapScreenUpdate ( void );
 extern	void			ChangeBitmapWindowX ( int );
@@ -31,3 +58,6 @@ extern	void			ChangeBitmapWindowY ( int );
 extern	byteval			BitsForColourInfo;
 extern	int				BitmapWindowX;
 extern	int				BitmapWindowY;
+extern	unsigned short	StartPosnX;
+
+#endif	/* BITMAP_H */
