@@ -1,7 +1,10 @@
 /*
- * Copyright (c) James Fidell 1994, 1995, 1996.
  *
- * Permission to use, copy, modify, distribute, and sell this software
+ * $Id: Memory.c,v 1.12 2002/01/15 15:46:43 james Exp $
+ *
+ * Copyright (C) James Fidell 1994-2002.
+ *
+ * Permission to use, copy, modify and distribute this software
  * and its documentation for any purpose is hereby granted without fee,
  * provided that the above copyright notice appear in all copies and
  * that both that copyright notice and this permission notice appear in
@@ -22,11 +25,66 @@
  *
  */
 
+/*
+ * Modification History
+ *
+ * $Log: Memory.c,v $
+ * Revision 1.12  2002/01/15 15:46:43  james
+ * *** empty log message ***
+ *
+ * Revision 1.11  2002/01/13 22:27:19  james
+ * Fix compile-time warnings
+ *
+ * Revision 1.10  2000/08/16 17:58:28  james
+ * Update copyright message
+ *
+ * Revision 1.9  1996/11/11 23:53:14  james
+ * Should set ScreenMemoryChanged, not ScreenImageChanged, when the screen
+ * memory is written.
+ *
+ * Revision 1.8  1996/10/12 15:15:04  james
+ * Compat.h added to support non-UNIXisms, in particular Win32.
+ *
+ * Revision 1.7  1996/10/10 21:58:37  james
+ * Cosmetic changes.
+ *
+ * Revision 1.6  1996/10/09 22:06:54  james
+ * Overhaul of the bitmapped screen handling code with particular respect to
+ * colour maps.
+ *
+ * Revision 1.5  1996/09/30 23:39:34  james
+ * Split out option processing into Options.[ch].  Updated the help message,
+ * added support for the Model A using the -a switch (and added the
+ * MODEL_B_ONLY #define in Config.h, added the -m and -s switches to set the
+ * initial screen mode and keyboard DIP switches.
+ *
+ * Revision 1.4  1996/09/24 23:05:39  james
+ * Update copyright dates.
+ *
+ * Revision 1.3  1996/09/23 16:09:51  james
+ * Initial implementation of bitmap MODEs -- including modification of
+ * screen handling to use different windows for teletext and bitmapped
+ * modes and corrections/improvements to colour- and cursor-handling
+ * code.
+ *
+ * Revision 1.2  1996/09/21 22:13:49  james
+ * Replaced "unsigned char" representation of 1 byte with "byteval".
+ *
+ * Revision 1.1  1996/09/21 17:20:38  james
+ * Source files moved to src directory.
+ *
+ * Revision 1.1.1.1  1996/09/21 13:52:48  james
+ * Xbeeb v0.1 initial release
+ *
+ *
+ */
+
 
 #include <stdio.h>
-#include <string.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <stdlib.h>
+#include <string.h>
 #include <limits.h>
 #include <memory.h>
 #include <sys/stat.h>
@@ -39,6 +97,8 @@
 #include "Sheila.h"
 #include "Screen.h"
 #include "Modes.h"
+#include "Compat.h"
+
 
 /*
  * Memory for direct access and for the screen change mirror
@@ -48,8 +108,7 @@
 byteval 			ScreenCheck [ 32768 ];
 byteval 			Mem [ 65536 ];
 
-
-#ifndef	MODEL_B_ONLY
+#ifndef   MODEL_B_ONLY
 
 /*
  * The highest RAM address -- only when Model A emulation is enabled
@@ -59,6 +118,7 @@ unsigned int		MaxRAMAddress = 0x7fff;
 
 #endif	/* MODEL_B_ONLY */
 
+
 /*
  * Memory from 0x8000 to 0xbfff -- language and utility ROMS and possibly
  * also some RAM.  There are potentially 16 of these slots, paged in on
@@ -69,7 +129,7 @@ unsigned int		MaxRAMAddress = 0x7fff;
  */
 
 byteval				PagedMem [ 16 ][ 16384 ];
-unsigned char		PageWrite [ 16 ] =
+byteval				PageWrite [ 16 ] =
 {
 	0, 0, 0, 0,
 	0, 0, 0, 0,
@@ -97,7 +157,7 @@ WriteByte ( unsigned int addr, byteval val )
 #ifdef	MODEL_B_ONLY
 		Mem [ addr ] = val;
 #else
-		Mem [ addr & MaxRAMAddress ] = val;
+		Mem [ addr & MaxRAMAddress  ] = val;
 #endif
 		if ( addr >= StartOfScreenMemory )
 		{
@@ -183,12 +243,12 @@ LoadOS ( char *osname )
 	 * ROM locations...
 	 */
 
-	if (( fd = open ( osname, O_RDONLY )) < 0 )
+	if (( fd = open ( osname, O_RDONLY | O_BINARY )) < 0 )
 	{
 		( void ) strcpy ( FnameBuffer, XBEEBROMS );
 		( void ) strcat ( FnameBuffer, osname );
 
-		if (( fd = open ( FnameBuffer, O_RDONLY )) < 0 )
+		if (( fd = open ( FnameBuffer, O_RDONLY | O_BINARY )) < 0 )
 		{
 			fprintf ( stderr, "Couldn't open OS ROM %s\n", osname );
 			exit ( 1 );
@@ -243,7 +303,7 @@ LoadPagedRom ( char *romname, byteval socket )
 		exit ( 1 );
 	}
 
-	if (( fd = open ( romname, O_RDONLY )) == 0 )
+	if (( fd = open ( romname, O_RDONLY | O_BINARY )) == 0 )
 	{
 		fprintf ( stderr, "Couldn't open paged ROM %s\n", romname );
 		exit ( 1 );
@@ -298,7 +358,7 @@ SaveUserRam ( char *fname )
 {
 	int				fd;
 
-	fd = open ( fname, O_WRONLY | O_CREAT | O_TRUNC, 0644 );
+	fd = open ( fname, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, 0644 );
 	write ( fd, ( void* ) Mem, 32768 );
 	close ( fd );
 }
