@@ -1,7 +1,10 @@
 /*
- * Copyright (c) James Fidell 1994.
  *
- * Permission to use, copy, modify, distribute, and sell this software
+ * $Id: Patchlevel.h,v 1.8 2002/01/15 15:46:43 james Exp $
+ *
+ * Copyright (C) James Fidell 1994-2002.
+ *
+ * Permission to use, copy, modify and distribute this software
  * and its documentation for any purpose is hereby granted without fee,
  * provided that the above copyright notice appear in all copies and
  * that both that copyright notice and this permission notice appear in
@@ -23,6 +26,11 @@
  */
 
 
+#ifndef	PATCHLEVEL_H
+#define	PATCHLEVEL_H
+
 #define	VERSION		0
-#define	RELEASE		2
+#define	RELEASE		4
 #define	PATCHLEVEL	0
+
+#endif	/* PATCHLEVEL_H */
