@@ -1,7 +1,10 @@
 /*
- * Copyright (c) James Fidell 1994, 1995, 1996.
  *
- * Permission to use, copy, modify, distribute, and sell this software
+ * $Id: Screen.c,v 1.24 2002/01/15 15:46:43 james Exp $
+ *
+ * Copyright (C) James Fidell 1994-2002.
+ *
+ * Permission to use, copy, modify and distribute this software
  * and its documentation for any purpose is hereby granted without fee,
  * provided that the above copyright notice appear in all copies and
  * that both that copyright notice and this permission notice appear in
@@ -22,9 +25,99 @@
  *
  */
 
+/*
+ * Modification History
+ *
+ * $Log: Screen.c,v $
+ * Revision 1.24  2002/01/15 15:46:43  james
+ * *** empty log message ***
+ *
+ * Revision 1.23  2002/01/14 22:18:51  james
+ * Added support for .inf files
+ *
+ * Revision 1.22  2002/01/13 23:29:46  james
+ * Unbreak MODE7 cursor
+ *
+ * Revision 1.21  2002/01/13 22:27:19  james
+ * Fix compile-time warnings
+ *
+ * Revision 1.20  2000/09/07 21:30:39  james
+ * Fix coredump
+ *
+ * Revision 1.19  2000/08/16 17:58:28  james
+ * Update copyright message
+ *
+ * Revision 1.18  2000/08/16 17:41:45  james
+ * Changes to work on TrueColor displays
+ *
+ * Revision 1.17  1996/11/11 23:52:31  james
+ * Corrections for non-MITSHM code.
+ *
+ * Revision 1.16  1996/10/13 22:17:18  james
+ * Helps to #include the file that TV_LINES is #defined in :-)
+ *
+ * Revision 1.15  1996/10/13 22:03:56  james
+ * Set up constant TV_LINES (=625) and used that instead of local magic numbers.
+ *
+ * Revision 1.14  1996/10/13 21:59:57  james
+ * Changed all window size/position parameters to #defined values.
+ *
+ * Revision 1.13  1996/10/12 14:30:25  james
+ * Corrections to MITSHM #ifdefs around code freeing up resources on shutdown.
+ *
+ * Revision 1.12  1996/10/09 23:19:10  james
+ * Added support for using the MIT X11 Shared Memory Extensions.
+ *
+ * Revision 1.11  1996/10/09 22:06:55  james
+ * Overhaul of the bitmapped screen handling code with particular respect to
+ * colour maps.
+ *
+ * Revision 1.10  1996/10/08 00:12:02  james
+ * Display the program name and version on the icon and title bar.
+ *
+ * Revision 1.9  1996/10/08 00:04:33  james
+ * Added InfoWindow to show LED status.  Also required addition of the
+ * SHIFTLOCK_SOUND_HACK to prevent the Shift Lock LED being light up
+ * whenever the sound buffer is full, which means that far too much
+ * time can be spent re-drawing the LED.
+ *
+ * Revision 1.8  1996/10/01 22:10:02  james
+ * Split keyboard handling into kEYMAP_STRICT and KEYMAP_LEGEND models.
+ *
+ * Revision 1.7  1996/09/24 23:05:42  james
+ * Update copyright dates.
+ *
+ * Revision 1.6  1996/09/23 16:09:52  james
+ * Initial implementation of bitmap MODEs -- including modification of
+ * screen handling to use different windows for teletext and bitmapped
+ * modes and corrections/improvements to colour- and cursor-handling
+ * code.
+ *
+ * Revision 1.5  1996/09/22 19:23:21  james
+ * Add the emulated filing system code.
+ *
+ * Revision 1.4  1996/09/21 23:16:13  james
+ * Loading of new X fonts for double height.  Unloading of all fonts.
+ *
+ * Revision 1.3  1996/09/21 22:39:53  james
+ * Improved handling of instruction disassembly.
+ *
+ * Revision 1.2  1996/09/21 22:13:49  james
+ * Replaced "unsigned char" representation of 1 byte with "byteval".
+ *
+ * Revision 1.1  1996/09/21 17:20:40  james
+ * Source files moved to src directory.
+ *
+ * Revision 1.1.1.1  1996/09/21 13:52:48  james
+ * Xbeeb v0.1 initial release
+ *
+ *
+ */
+
 
 #include <stdio.h>
 #include <string.h>
+#include <stdlib.h>
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/keysym.h>
@@ -32,6 +125,7 @@
 #include "Config.h"
 #include "Beeb.h"
 #include "Screen.h"
+#include "Crtc.h"
 #include "Modes.h"
 #include "Teletext.h"
 #include "Bitmap.h"
@@ -39,14 +133,24 @@
 #include "Keyboard.h"
 #include "Keymap.h"
 #include "Memory.h"
-#include "EFS.h"
 #include "Patchlevel.h"
 
+#ifdef	EMUL_FS
+#include "EmulFS.h"
+#endif
+
+#ifdef	DGA
+#include <sys/types.h>
+#include <sys/wait.h>
+#include <X11/extensions/xf86dga.h>
+#include <X11/extensions/xf86vmode.h>
+#else	/* DGA */
 #ifdef	MITSHM
 #include <sys/ipc.h>
 #include <sys/shm.h>
 #include <X11/extensions/XShm.h>
 #endif	/* MITSHM */
+#endif	/* DGA */
 
 
 Display				*dpy;
@@ -85,7 +189,6 @@ GC					CopyAreaGC;
 Window				BitmapScreen;
 Pixmap				BitmapPixmap;
 int					BytesPerImageLine;
-char				*ImageData;
 
 /*
  * General colour-handling stuff
@@ -110,6 +213,10 @@ int					RgbValues [ 8 ][ 3 ] =
 	{ 0xffff, 0xffff, 0xffff }		/* 7 = White */
 };
 
+static const char	*Colours[8] = {
+	"black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"
+};
+
 
 static unsigned int	ScreenLengths [ 4 ] =
 {
@@ -139,26 +246,27 @@ static XShmSegmentInfo	SharedSegInfo;
 void
 InitialiseScreen()
 {
-	int					i, dummy;
+	int					i;
 	XColor				colour, scol, xcol;
 	Window				Root;
 	XGCValues			CursorGCValues;
-	unsigned long		PlaneMask;
-	int					DefScreen, DefDepth;
+	unsigned long		PlaneMask = ~0;
+	int					DefScreen, DefDepth, visuals, VisualClass;
 	Visual				*DefVisual;
+	XVisualInfo			VisualTmpl, *MyVisual;
 	GC					DefGC;
 	XWindowChanges		Raise;
 	char				TitleString [ 80 ];
 	XSizeHints			SizeHints;
 #ifdef	MITSHM
-	int					PixmapFormat;
+	int					PixmapFormat, dummy;
 	Bool				SharedPixmapSupport;
 #endif
 
 
 	if (( dpy = XOpenDisplay ( 0 )) == 0 )
 	{
-		fprintf ( stderr, "Couldn't open connection to dispay\n" );
+		fprintf ( stderr, "Couldn't open connection to display\n" );
 		exit ( 1 );
 	}
 
@@ -168,6 +276,24 @@ InitialiseScreen()
 	DefDepth = DefaultDepth ( dpy, DefScreen );
 	DefGC = DefaultGC ( dpy, DefScreen );
 
+	if ( DefDepth < 3 ) {
+		fprintf ( stderr, "Screen depth of %d is not supported\n", DefDepth );
+		exit ( 1 );
+	}
+
+	VisualTmpl.visual = DefVisual;
+	VisualTmpl.screen = DefScreen;
+	VisualTmpl.depth = DefDepth;
+	MyVisual = XGetVisualInfo ( dpy, VisualScreenMask | VisualDepthMask,
+		&VisualTmpl, &visuals );
+	if ( visuals != 1 ) {
+		fprintf ( stderr, "can't handle default screen with multiple "
+			"visuals\n" );
+		exit ( 1 );
+	}
+	VisualClass = MyVisual->class;
+	XFree ( MyVisual );
+
 	/*
 	 * FIX ME
 	 *
@@ -181,32 +307,47 @@ InitialiseScreen()
 	InfoWindowBlack = BlackPixel ( dpy, DefScreen );
 	InfoWindowWhite = WhitePixel ( dpy, DefScreen );
 
-	/*
-	 * Get four planes in the colourmap
-	 */
+	if ( VisualClass == PseudoColor ) {
+		/*
+	 	* Get four planes in the colourmap
+	 	*/
 
-	if ( XAllocColorCells( dpy, DefCmap, False, Masks, 4, &ColourBits, 1 ) == 0)
-	{
-		fprintf ( stderr, "Failed to allocate colour planes\n" );
+		if ( XAllocColorCells( dpy, DefCmap, False, Masks, 4,
+			&ColourBits, 1 ) == 0)
+		{
+			fprintf ( stderr, "Failed to allocate colour planes\n" );
+			exit ( 1 );
+		}
+
+		for ( i = 0; i < 16; i++ )
+		{
+			Cells [ i ] = ColourBits | ( i & 1 ? Masks [ 0 ] : 0 ) |
+				( i & 2 ? Masks [ 1 ] : 0 ) | ( i & 4 ? Masks [ 2 ] : 0 ) |
+				( i & 8 ? Masks [ 3 ] : 0 );
+			colour.pixel = Cells [ i ];
+			colour.red = RgbValues [ i % 8 ][ 0 ];
+			colour.green = RgbValues [ i % 8 ][ 1 ];
+			colour.blue = RgbValues [ i % 8 ][ 2 ];
+			colour.flags = DoRed | DoGreen | DoBlue;
+			XStoreColor ( dpy, DefCmap, &colour );
+		}
+
+		PlaneMask = ( Masks [ 0 ] | Masks [ 1 ] | Masks [ 2 ] | Masks [ 3 ] ) &
+					~( Masks [ 0 ] & Masks [ 1 ] & Masks [ 2 ] & Masks [ 4 ] );
+	} else if ( VisualClass == TrueColor ) {
+		Cells[0] = BlackPixel ( dpy, DefScreen );
+		Cells[7] = WhitePixel ( dpy, DefScreen );
+		Cells[8] = BlackPixel ( dpy, DefScreen );
+		Cells[15] = WhitePixel ( dpy, DefScreen );
+		for ( i = 1; i < 7; i++ ) {
+			XAllocNamedColor ( dpy, DefCmap, Colours[i], &scol, &xcol );
+			Cells[i] = Cells[i+8] = scol.pixel;
+		}
+	} else {
+		fprintf ( stderr, "Can't handle visual class %d\n", VisualClass );
 		exit ( 1 );
 	}
 
-	for ( i = 0; i < 16; i++ )
-	{
-		Cells [ i ] = ColourBits | ( i & 1 ? Masks [ 0 ] : 0 ) |
-					( i & 2 ? Masks [ 1 ] : 0 ) | ( i & 4 ? Masks [ 2 ] : 0 ) |
-												( i & 8 ? Masks [ 3 ] : 0 );
-		colour.pixel = Cells [ i ];
-		colour.red = RgbValues [ i % 8 ][ 0 ];
-		colour.green = RgbValues [ i % 8 ][ 1 ];
-		colour.blue = RgbValues [ i % 8 ][ 2 ];
-		colour.flags = DoRed | DoGreen | DoBlue;
-		XStoreColor ( dpy, DefCmap, &colour );
-	}
-
-	PlaneMask = ( Masks [ 0 ] | Masks [ 1 ] | Masks [ 2 ] | Masks [ 3 ] ) &
-				~( Masks [ 0 ] & Masks [ 1 ] & Masks [ 2 ] & Masks [ 4 ] );
-
 	/*
 	 * FIX ME
 	 *
@@ -216,20 +357,20 @@ InitialiseScreen()
 	 * directly.
 	 */
 
-	BeebScreen = XCreateSimpleWindow ( dpy, Root, 0, 0, 640, 566, 0,
-															White, Black );
-	InfoWindow = XCreateSimpleWindow ( dpy, BeebScreen, 0, 512, 640, 40, 0,
-															White, Black );
+	BeebScreen = XCreateSimpleWindow ( dpy, Root, 0, 0, MAIN_WIN_W,
+											MAIN_WIN_H, 0, White, Black );
+	InfoWindow = XCreateSimpleWindow ( dpy, BeebScreen, INFO_WIN_X,
+					INFO_WIN_Y, INFO_WIN_W, INFO_WIN_H, 0, White, Black );
 	TeletextScreen = XCreateSimpleWindow ( dpy, BeebScreen, TeletextWindowX,
-								TeletextWindowY, 480, 475, 0, White, Black );
+					TeletextWindowY, TTXT_WIN_W, TTXT_WIN_H, 0, White, Black );
 	BitmapScreen = XCreateSimpleWindow ( dpy, BeebScreen, BitmapWindowX,
-								BitmapWindowY, 640, 512, 0, White, Black );
+					BitmapWindowY, BMAP_WIN_W, BMAP_WIN_H, 0, White, Black );
 
 	(void) sprintf ( TitleString, TITLE_STRING, VERSION, RELEASE, PATCHLEVEL );
 
 	SizeHints.flags = PSize;
-	SizeHints.width = 640;
-	SizeHints.height = 566;
+	SizeHints.width = MAIN_WIN_W;
+	SizeHints.height = MAIN_WIN_H;
 
 	XSetStandardProperties ( dpy, BeebScreen, TitleString, TitleString, None,
 														0, 0, &SizeHints );
@@ -253,7 +394,7 @@ InitialiseScreen()
 			 */
 
 			BitmapImage = XShmCreateImage ( dpy, DefVisual, DefDepth, ZPixmap,
-				0, &SharedSegInfo, 640, 624 );
+				0, &SharedSegInfo, BMAP_WIN_W, TV_LINES );
 
 			SharedSegInfo.shmid = shmget ( IPC_PRIVATE,
 				BitmapImage -> bytes_per_line * BitmapImage -> height,
@@ -276,9 +417,9 @@ InitialiseScreen()
 			 */
 
 			BitmapPixmap = XShmCreatePixmap ( dpy, BitmapScreen,
-				SharedSegInfo.shmaddr, &SharedSegInfo, 640, 624, DefDepth );
+				SharedSegInfo.shmaddr, &SharedSegInfo, BMAP_WIN_W,
+														TV_LINES, DefDepth );
 
-			ImageData = BitmapImage -> data;
 			BytesPerImageLine = BitmapImage -> bytes_per_line;
 		}
 		else
@@ -300,7 +441,8 @@ InitialiseScreen()
 	 * is just the bit that can be seen on the display.
 	 */
 
-	BitmapPixmap = XCreatePixmap ( dpy, BitmapScreen, 640, 624, DefDepth );
+	BitmapPixmap = XCreatePixmap ( dpy, BitmapScreen, BMAP_WIN_W,
+													BMAP_WIN_H, DefDepth );
 	
 
 #ifdef	MITSHM
@@ -407,7 +549,8 @@ InitialiseScreen()
 	 * Now clear the main Beeb window and the bitmapped screen to black.
 	 */
 
-	XFillRectangle ( dpy, BitmapPixmap, BitmapGC [ 0 ], 0, 0, 639, 623 );
+	XFillRectangle ( dpy, BitmapPixmap, BitmapGC [ 0 ], 0, 0, MAIN_WIN_W - 1,
+															MAIN_WIN_H - 1 );
 	XClearWindow ( dpy, BeebScreen );
 	XClearWindow ( dpy, InfoWindow );
 
@@ -431,12 +574,16 @@ ShutdownScreen()
 	XUnloadFont ( dpy, TtextSeparateMosaicDblU );
 	XUnloadFont ( dpy, TtextSeparateMosaicDblL );
 
+#ifdef	MITSHM
 	XShmDetach ( dpy, &SharedSegInfo );
+#endif
 	XFreePixmap ( dpy, BitmapPixmap );
+#ifdef	MITSHM
 	shmdt ( SharedSegInfo.shmaddr );
 	shmctl ( SharedSegInfo.shmid, IPC_RMID, 0 );
 
 	XDestroyImage ( BitmapImage );
+#endif
 
 	/*
 	 * FIX ME
