@@ -1,7 +1,10 @@
 /*
- * Copyright (c) James Fidell 1994, 1995, 1996.
  *
- * Permission to use, copy, modify, distribute, and sell this software
+ * $Id: SystemVia.c,v 1.14 2002/01/15 15:46:43 james Exp $
+ *
+ * Copyright (C) James Fidell 1994-2002.
+ *
+ * Permission to use, copy, modify and distribute this software
  * and its documentation for any purpose is hereby granted without fee,
  * provided that the above copyright notice appear in all copies and
  * that both that copyright notice and this permission notice appear in
@@ -22,9 +25,97 @@
  *
  */
 
+/*
+ * Modification History
+ *
+ * $Log: SystemVia.c,v $
+ * Revision 1.14  2002/01/15 15:46:43  james
+ * *** empty log message ***
+ *
+ * Revision 1.13  2002/01/13 22:27:19  james
+ * Fix compile-time warnings
+ *
+ * Revision 1.12  2000/08/16 17:58:28  james
+ * Update copyright message
+ *
+ * Revision 1.11  1996/11/24 22:13:27  james
+ * Timer values need to be updated before they are read or over-written
+ * in the 6522 User and System VIA code.  From a fix by David Ralph Stacey.
+ *
+ * Revision 1.10  1996/11/19 00:56:07  james
+ * Writing to T2CL should clear any outstanding T2 interrupt.  Fix from
+ * David Ralph Stacey.
+ *
+ * Revision 1.9  1996/11/15 08:50:57  james
+ * When the IER is written, disabled interrupts need to be cleared as well
+ * as setting newly enabled interrupts.  (Fix from David Ralph Stacey.)
+ *
+ * Revision 1.8  1996/10/01 00:33:03  james
+ * Created separate hardware reset code for each emulated unit and called
+ * these from the main initialisation section of the code to do all of the
+ * setup necessary.
+ *
+ * Revision 1.7  1996/09/25 19:19:57  james
+ * Major overhaul of VIA emulation code :
+ *
+ *   Enabled toggling of PB7 in system VIA depending on ACR bit 6 and the
+ *   one-shot/free-run mode of T1
+ *
+ *   Implemented User VIA T1 free-running mode.  Set the initial value of
+ *   the User VIA ORA to 0x80.  Planetoid/Defender now works for the first
+ *   time!
+ *
+ *   Corrected value returned by read from VIA T2CL and T2CH.  Frak! now
+ *   works.
+ *
+ *   Set up dummy return for reads from the system VIA IRA and SR.
+ *
+ *   Implemented address wrap-around for memory-mapped registers in the VIA.
+ *
+ *   Set up dummy return for reads from the user VIA SR.
+ *
+ *   Implemented address wrap-around for memory-mapped registers in the VIA.
+ *
+ *   Updated 6522 VIA emulation to have correct initial values for VIA
+ *   registers wherever possible.
+ *
+ *   Heavily modified 6522 VIA code to separate out the input/output
+ *   registers and what is actually on the data pins.  This has the benefits
+ *   of tidying up the whole VIA i/o emulation and not requiring any nasty
+ *   configuration hacks to get software to work (apart from those that exist
+ *   because of uncompleted emulation).
+ *
+ *   Tidied up 6522Via interrupt handling code.
+ *
+ * Revision 1.6  1996/09/24 23:05:44  james
+ * Update copyright dates.
+ *
+ * Revision 1.5  1996/09/23 19:02:13  james
+ * Major overhaul of the VIA emulation code.
+ *
+ * Revision 1.4  1996/09/22 21:00:54  james
+ * Changed IRQ-handling code to be called only when I think an IRQ may have
+ * happened.  Converted it all to #defines so that it can be compiled inline.
+ *
+ * Revision 1.3  1996/09/22 20:36:13  james
+ * More conversions of exit(x) to FatalError()
+ *
+ * Revision 1.2  1996/09/21 22:13:51  james
+ * Replaced "unsigned char" representation of 1 byte with "byteval".
+ *
+ * Revision 1.1  1996/09/21 17:20:41  james
+ * Source files moved to src directory.
+ *
+ * Revision 1.1.1.1  1996/09/21 13:52:48  james
+ * Xbeeb v0.1 initial release
+ *
+ *
+ */
+
 
 #include <stdio.h>
 #include <unistd.h>
+#include <string.h>
 
 #include "Config.h"
 #include "6502.h"
@@ -53,7 +144,13 @@ static unsigned char	PortALatchEnable;
 static unsigned char	PortBLatchEnable;
 static byteval			CA1Control, CA2Control, CB1Control, CB2Control;
 
-#define	DEFAULT_PORT_B	0xf0;
+
+/*
+ * This is a hack to cover bits of the emulation that doesn't exist yet.
+ * See where it's used in the code for more detail.
+ */
+
+#define	DEFAULT_PORT_B	0xf0
 
 
 void
@@ -277,11 +374,19 @@ ReadSystemVia ( int addr )
 			break;
 
 		case T1CL :
+#ifdef	FASTCLOCK
+			ViaClockUpdate ( ClockCyclesSoFar );
+			ClockCyclesSoFar = 0;
+#endif	/* FASTCLOCK */
 			SystemViaClearInterrupt ( INT_T1 );
 			return SystemViaTimer1 & 0xff;
 			break;
 
 		case T1CH :
+#ifdef	FASTCLOCK
+			ViaClockUpdate ( ClockCyclesSoFar );
+			ClockCyclesSoFar = 0;
+#endif	/* FASTCLOCK */
 			return SystemViaTimer1 >> 8;
 			break;
 
@@ -294,11 +399,20 @@ ReadSystemVia ( int addr )
 			break;
 
 		case T2CL :
+#ifdef	FASTCLOCK
+			ViaClockUpdate ( ClockCyclesSoFar );
+			ClockCyclesSoFar = 0;
+#endif	/* FASTCLOCK */
 			SystemViaTimer2InterruptEnable = 1;
+			SystemViaClearInterrupt ( INT_T2 );
 			return SystemViaTimer2 & 0xff;
 			break;
 
 		case T2CH :
+#ifdef	FASTCLOCK
+			ViaClockUpdate ( ClockCyclesSoFar );
+			ClockCyclesSoFar = 0;
+#endif	/* FASTCLOCK */
 			return SystemViaTimer2 >> 8;
 			break;
 
@@ -599,6 +713,10 @@ WriteSystemVia ( int addr, byteval val )
 			 * output
 			 */
 
+#ifdef	FASTCLOCK
+			ViaClockUpdate ( ClockCyclesSoFar );
+			ClockCyclesSoFar = 0;
+#endif	/* FASTCLOCK */
 			SystemVia [ T1CH ] = SystemVia [ T1LH ] = val;
 			SystemVia [ T1CL ] = SystemVia [ T1LL ];
 			SystemViaTimer1 = SystemVia [ T1CH ] * 256 + SystemVia [ T1CL ];
@@ -624,6 +742,10 @@ WriteSystemVia ( int addr, byteval val )
 			 * interrupt flag in the IFR.
 			 */
 
+#ifdef	FASTCLOCK
+			ViaClockUpdate ( ClockCyclesSoFar );
+			ClockCyclesSoFar = 0;
+#endif	/* FASTCLOCK */
 			SystemVia [ T2CH ] = val;
 			SystemVia [ T2CL ] = SystemVia [ T2LL ];
 			SystemViaTimer2 = SystemVia [ T2CH ] * 256 + SystemVia [ T2CL ];
@@ -690,21 +812,23 @@ WriteSystemVia ( int addr, byteval val )
 			 * If bit 7 is set, then each set bit in sets the corresponding
 			 * bit in the IER.
 			 *
+			 * Having set up new flags, it may be required that we set/clear
+			 * some of the interrupts in the IFR.
 			 */
 
 			if ( val & 0x80 )
+			{
 				SystemVia [ IER ] |= ( val & 0x7f );
+				SystemViaSetInterrupt ( SystemVia [ IFR ] );
+			}
 			else
+			{
 				SystemVia [ IER ] &= ( val ^ 0x7f );
+				SystemViaClearInterrupt ( val );
+			}
 #ifdef	INFO
 			printf ( "System VIA IER = %02x\n", SystemVia [ IER ] );
 #endif
-
-			/*
-			 * Now check to see if we need to cause an IRQ.
-			 */
-
-			SystemViaSetInterrupt ( SystemVia [ IFR ] );
 			break;
 
 		case IFR :
@@ -864,6 +988,10 @@ SaveSystemVia ( int fd )
 {
 	byteval		via [ 32 ];
 
+#ifdef	FASTCLOCK
+	ViaClockUpdate ( ClockCyclesSoFar );
+	ClockCyclesSoFar = 0;
+#endif	/* FASTCLOCK */
 	SystemVia [ T1CL ] = SystemViaTimer1 & 0xff;
 	SystemVia [ T1CH ] = SystemViaTimer1 >> 8;
 	SystemVia [ T2CL ] = SystemViaTimer2 & 0xff;
