@@ -1,7 +1,10 @@
 /*
- * Copyright (c) James Fidell 1994.
  *
- * Permission to use, copy, modify, distribute, and sell this software
+ * $Id: Teletext.c,v 1.20 2002/01/15 15:46:43 james Exp $
+ *
+ * Copyright (C) James Fidell 1994-2002.
+ *
+ * Permission to use, copy, modify and distribute this software
  * and its documentation for any purpose is hereby granted without fee,
  * provided that the above copyright notice appear in all copies and
  * that both that copyright notice and this permission notice appear in
@@ -22,12 +25,96 @@
  *
  */
 
+/*
+ * Modification History
+ *
+ * $Log: Teletext.c,v $
+ * Revision 1.20  2002/01/15 15:46:43  james
+ * *** empty log message ***
+ *
+ * Revision 1.19  2000/08/16 17:58:28  james
+ * Update copyright message
+ *
+ * Revision 1.18  2000/08/16 17:41:45  james
+ * Changes to work on TrueColor displays
+ *
+ * Revision 1.17  1996/11/25 00:01:19  james
+ * Remove the "Default Vertical Sync value" magic number.
+ *
+ * Revision 1.16  1996/11/24 21:54:10  james
+ * Change MIN to XBEEB_MIN to avoid clashes with some OSes.
+ *
+ * Revision 1.15  1996/11/09 23:02:31  james
+ * Take account of the horizontal start of display based on the clock rate
+ * set in the Video ULA and the CRTC Horiz. Sync. register.
+ *
+ * Revision 1.14  1996/10/13 21:59:59  james
+ * Changed all window size/position parameters to #defined values.
+ *
+ * Revision 1.13  1996/10/13 16:38:32  james
+ * Screen addresses need to be recalculated on switching between bitmapped
+ * and non-bitmapped modes because they're calculated differently from the
+ * values in the registers.
+ *
+ * Revision 1.12  1996/10/13 12:13:33  james
+ * Parenthesise all parameters to #defined macros.
+ *
+ * Revision 1.11  1996/10/10 22:09:02  james
+ * HoldCharacter needs to change back to space for TT_DOUBLE_HEIGHT
+ *
+ * Revision 1.10  1996/10/09 22:06:57  james
+ * Overhaul of the bitmapped screen handling code with particular respect to
+ * colour maps.
+ *
+ * Revision 1.9  1996/10/08 00:04:35  james
+ * Added InfoWindow to show LED status.  Also required addition of the
+ * SHIFTLOCK_SOUND_HACK to prevent the Shift Lock LED being light up
+ * whenever the sound buffer is full, which means that far too much
+ * time can be spent re-drawing the LED.
+ *
+ * Revision 1.8  1996/09/30 23:39:35  james
+ * Split out option processing into Options.[ch].  Updated the help message,
+ * added support for the Model A using the -a switch (and added the
+ * MODEL_B_ONLY #define in Config.h, added the -m and -s switches to set the
+ * initial screen mode and keyboard DIP switches.
+ *
+ * Revision 1.7  1996/09/24 23:05:44  james
+ * Update copyright dates.
+ *
+ * Revision 1.6  1996/09/23 16:09:52  james
+ * Initial implementation of bitmap MODEs -- including modification of
+ * screen handling to use different windows for teletext and bitmapped
+ * modes and corrections/improvements to colour- and cursor-handling
+ * code.
+ *
+ * Revision 1.5  1996/09/22 22:34:33  james
+ * Completed handling of teletext HOLD mode.
+ *
+ * Revision 1.4  1996/09/22 22:21:09  james
+ * Add handling of double-height fonts.
+ *
+ * Revision 1.3  1996/09/22 22:01:53  james
+ * Use #defined values for teletext control characters.
+ *
+ * Revision 1.2  1996/09/21 18:18:18  james
+ * Corrections to variable typing for CursorX, CursorY and line.
+ *
+ * Revision 1.1  1996/09/21 17:20:42  james
+ * Source files moved to src directory.
+ *
+ * Revision 1.1.1.1  1996/09/21 13:52:48  james
+ * Xbeeb v0.1 initial release
+ *
+ *
+ */
+
 
 #include <stdio.h>
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 
 #include "Config.h"
+#include "Beeb.h"
 #include "Teletext.h"
 #include "Memory.h"
 #include "Screen.h"
@@ -86,9 +173,8 @@
 #define	DH_UPPER		1
 #define	DH_LOWER		2
 
-
-int						TeletextWindowX = 80;
-int						TeletextWindowY = 0;
+int						TeletextWindowX = TTXT_WIN_X;
+int						TeletextWindowY = TTXT_WIN_Y;
 
 static int				FlashTimer;
 static int				CursorTimer;
@@ -104,40 +190,56 @@ static unsigned int		CursorViewable = 0;
 
 static void				DrawCursor();
 
-#define	MIN(x,y)	(( x < y ) ? x : y )
 
 void
 InitialiseTeletext()
 {
 	if ( CurrentScreenMode != MODE_TELETEXT )
 	{
-		int				i;
-		XColor			colour;
-
 		CurrentScreenMode = MODE_TELETEXT;
 
 		/*
-		 * FIX ME
-	 	 *
-		 * I'm not actually sure I need to do this...
-		 *
-		 * Set the first 8 entries into the colourmap back to their
-		 * default values -- they might have been changed if we've
-		 * changed from a graphics mode where the colours were
-		 * redefined.
+		 * Need to recalculate the top of screen information here because
+		 * it's not going to be right given that teletext screen start
+		 * addresses are calculated differently to bitmap screen start
+		 * addresses.  The same applies to the cursor address where more
+		 * calculations have to be done...
 		 */
 
-		for ( i = 0; i < 8; i++ )
-		{
-			Cells [ i ] = ColourBits | ( i & 1 ? Masks [ 0 ] : 0 ) |
-					( i & 2 ? Masks [ 1 ] : 0 ) | ( i & 4 ? Masks [ 2 ] : 0 );
-			colour.pixel = Cells [ i ];
-			colour.red = RgbValues [ i % 8 ][ 0 ];
-			colour.green = RgbValues [ i % 8 ][ 1 ];
-			colour.blue = RgbValues [ i % 8 ][ 2 ];
-			colour.flags = DoRed | DoGreen | DoBlue;
-			XStoreColor ( dpy, Cmap, &colour );
-		}
+		TopOfScreen = (( ScreenStartHi ^ 0x20 ) + 0x74 ) << 8;
+		TopOfScreen |= ScreenStartLo;
+
+		/*
+		 * FIX ME
+		 *
+		 * I'm sure there's going to be a problem here if the cursor
+		 * drawing code doesn't realise that the cursor won't be on
+		 * the screen when the window pops up.
+		 */
+
+		CursorAddress = (( CursorPosHi ^ 0x20 ) + 0x74 ) << 8;
+		CursorAddress |= CursorPosLo;
+		CursorOffset = CursorAddress - TopOfScreen;
+
+		if ( CursorAddress < TopOfScreen )
+			CursorOffset += ScreenLength;
+
+		/*
+		 * FIX ME
+		 *
+		 * This is not right, because it doesn't take
+		 * account of the horizontal displayed value.
+		 *
+		 * It would be better to use
+		 * HorizDisplayed / ( 6845 chars per real char )
+		 * rather than CharsPerLine.
+		 *
+		 * This code should be corrected at the same time as the code
+		 * in Crtc.c (where I pinched it from).
+		 */
+
+		NewCursorX = CursorOffset % CharsPerLine;
+		NewCursorY = CursorOffset / CharsPerLine;
 
 		/*
 		 * FIX ME
@@ -150,6 +252,14 @@ InitialiseTeletext()
 		XMapRaised ( dpy, TeletextScreen );
 		XFlush ( dpy );
 	}
+
+	/*
+	 * Default vertical sync. position (so we put the window somewhere
+	 * sane on the screen...
+	 */
+
+	DefaultVertSync = 28;
+
 	FlashTimer = CursorTimer = 0;
 	FlashStatus = CursorOnScreen = 0;
 
@@ -206,11 +316,17 @@ TeletextScreenUpdate()
 
 	CursorTimer = ( CursorTimer + 1 ) % CursorBlinkFrequency;
 
-	if ( ScreenChanged || DoFlash )
+	if ( ScreenMemoryChanged || ScreenImageChanged || DoFlash )
 	{
 		SeenFlash = 0;
 		CursorOnScreen = 0;
 
+		/*
+		 * FIX ME
+		 *
+		 * We really should take account of StartPosnX here
+		 */
+
 		p = TopOfScreen;
 		for ( lines = 1; lines <= VertDisplayed; lines++ )
 		{
@@ -236,7 +352,7 @@ TeletextScreenUpdate()
 			/*
 			 * This is why DH_xxx must be set to specific values
 			 */
-
+ 
 			if ( DblHeight )
 				DblHeight = ( ++DblHeight ) % 3;
 
@@ -249,7 +365,12 @@ TeletextScreenUpdate()
 				 * but speed is of the essence
 				 */
 
+#ifdef	MODEL_B_ONLY
 				c = Mem [ p++ ];
+#else
+				c = Mem [ p & MaxRAMAddress ];
+				p++;
+#endif
 
 				/*
 				 * Non-displayable characters (I think!)
@@ -281,6 +402,9 @@ TeletextScreenUpdate()
 				/*
 				 * New text colour.
 				 *
+				 * FIX ME
+				 *
+				 * should this character look like the current hold character ?
 				 */
 
 				if ( c < TT_STEADY )
@@ -305,10 +429,7 @@ TeletextScreenUpdate()
 
 				if ( c == TT_STEADY )
 				{
-					if ( HoldMode )
-						line [ got++ ] = HoldCharacter;
-					else
-						line [ got++ ] = ' ';
+					line [ got++ ] = HoldMode ? HoldCharacter : ' ';
 					XDrawImageString ( dpy, TeletextScreen, CurrGC, hpos * 12,
 											lines * 19, (char *)line, got );
 					hpos += got;
@@ -324,10 +445,7 @@ TeletextScreenUpdate()
 
 				if ( c == TT_FLASH )
 				{
-					if ( HoldMode )
-						line [ got++ ] = HoldCharacter;
-					else
-						line [ got++ ] = ' ';
+					line [ got++ ] = HoldMode ? HoldCharacter : ' ';
 					XDrawImageString ( dpy, TeletextScreen, CurrGC, hpos * 12,
 											lines * 19, (char *)line, got );
 					hpos += got;
@@ -345,18 +463,16 @@ TeletextScreenUpdate()
 				 *
 				 * I'm sure the handling of this isn't quite right, but it's
 				 * reasonably close for the moment.
-				 *
+				 * 
 				 * Problems will, I believe, occur if the teletext DH/SH
 				 * control codes do not appear in the same position on
 				 * the second line.
+				 *
 				 */
 
 				if ( c == TT_NORMAL_HEIGHT )
 				{
-					if ( HoldMode )
-						line [ got++ ] = HoldCharacter;
-					else
-						line [ got++ ] = ' ';
+					line [ got++ ] = HoldMode ? HoldCharacter : ' ';
 					XDrawImageString ( dpy, TeletextScreen, CurrGC, hpos * 12,
 											lines * 19, (char *)line, got );
 					hpos += got;
@@ -411,11 +527,16 @@ TeletextScreenUpdate()
 							XSetFont ( dpy, TtextMosaicGC,
 											TtextSeparateMosaicDblU );
 					}
+
 					goto next_char;
 				}
 
 				/*
 				 * change graphics colour
+				 *
+				 * FIX ME
+				 *
+				 * Does the colour change before or after the hold character ?
 				 */
 
 				if ( c < TT_CONCEAL )
@@ -446,7 +567,6 @@ TeletextScreenUpdate()
 
 				/*
 				 * conceal display.
-				 *
 				 */
 
 				if ( c == TT_CONCEAL )
@@ -467,10 +587,7 @@ TeletextScreenUpdate()
 
 				if ( c == TT_CONTIGUOUS )
 				{
-					if ( HoldMode )
-						line [ got++ ] = HoldCharacter;
-					else
-						line [ got++ ] = ' ';
+					line [ got++ ] = HoldMode ? HoldCharacter : ' ';
 					XDrawImageString ( dpy, TeletextScreen, CurrGC, hpos * 12,
 											lines * 19, (char *)line, got );
 					XSetFont ( dpy, TtextMosaicGC, TtextContiguousMosaic );
@@ -485,10 +602,7 @@ TeletextScreenUpdate()
 
 				if ( c == TT_SEPARATED )
 				{
-					if ( HoldMode )
-						line [ got++ ] = HoldCharacter;
-					else
-						line [ got++ ] = ' ';
+					line [ got++ ] = HoldMode ? HoldCharacter : ' ';
 					XDrawImageString ( dpy, TeletextScreen, CurrGC, hpos * 12,
 											lines * 19, (char *)line, got );
 					XSetFont ( dpy, TtextMosaicGC, TtextSeparateMosaic );
@@ -553,10 +667,12 @@ TeletextScreenUpdate()
 					 * we're currently in graphics mode
 					 */
 
-					if ( CurrGC != TtextTextGC )
-						HoldCharacter = Mem [ p - 2 ];
-					else
-						HoldCharacter = ' ';
+					HoldCharacter = ( CurrGC == TtextTextGC ) ? ' ' :
+#ifdef	MODEL_B_ONLY
+											Mem [ p - 2 ];
+#else
+											Mem [ ( p  & MaxRAMAddress ) - 2 ];
+#endif
 
 					line [ got++ ] = HoldCharacter;
 					goto next_char;
@@ -600,7 +716,7 @@ next_char:
 
 	DrawCursor();
 
-	ScreenChanged = 0;
+	ScreenImageChanged = ScreenMemoryChanged = 0;
 	XFlush ( dpy );
 	return;
 }
@@ -640,13 +756,14 @@ DrawCursor()
 				if ( CursorResized )
 				{
 					if ( CursorStartLine > 19 || ( CursorByteWidth == 0 &&
-						 MasterCursorWidth == 0 ) || CursorEndLine <
-														CursorStartLine )
+							MasterCursorWidth == 0 ) || CursorEndLine <
+															CursorStartLine )
 						CursorViewable = 0;
 					else
 						CursorViewable = 1;
 					CursorY = NewCursorY * 19 + CursorStartLine;
-					CursorDepth = MIN(19, CursorEndLine - CursorStartLine + 1);
+					CursorDepth = XBEEB_MIN ( 19, ( CursorEndLine -
+														CursorStartLine + 1 ));
 					CursorWidth = ( CursorByteWidth - 1 + MasterCursorWidth )
 																		* 12;
 					CursorResized = 0;
@@ -670,13 +787,14 @@ DrawCursor()
 				if ( CursorResized )
 				{
 					if ( CursorStartLine > 19 || ( CursorByteWidth == 0 &&
-						 MasterCursorWidth == 0 ) || CursorEndLine <
-														CursorStartLine )
+							MasterCursorWidth == 0 ) || CursorEndLine <
+															CursorStartLine )
 						CursorViewable = 0;
 					else
 						CursorViewable = 1;
 					CursorY = NewCursorY * 19 + CursorStartLine;
-					CursorDepth = MIN(19, CursorEndLine - CursorStartLine + 1);
+					CursorDepth = XBEEB_MIN ( 19, ( CursorEndLine -
+														CursorStartLine + 1 ));
 					CursorWidth = ( CursorByteWidth - 1 + MasterCursorWidth )
 																		* 12;
 					CursorResized = 0;
@@ -700,7 +818,7 @@ DrawCursor()
 				{
 					if ( CursorViewable )
 						XFillRectangle ( dpy, TeletextScreen, CursorGC,
-							CursorX, CursorY, CursorWidth, CursorDepth );
+								CursorX, CursorY, CursorWidth, CursorDepth );
 
 					if ( CursorMoved )
 					{
@@ -712,14 +830,14 @@ DrawCursor()
 					if ( CursorResized )
 					{
 						if ( CursorStartLine > 19 || ( CursorByteWidth == 0 &&
-						 MasterCursorWidth == 0 ) || CursorEndLine <
-														CursorStartLine )
+							MasterCursorWidth == 0 ) || CursorEndLine <
+															CursorStartLine )
 							CursorViewable = 0;
 						else
 							CursorViewable = 1;
 						CursorY = NewCursorY * 19 + CursorStartLine;
-						CursorDepth = MIN ( 19, CursorEndLine -
-														CursorStartLine + 1 );
+						CursorDepth = XBEEB_MIN ( 19, ( CursorEndLine -
+														CursorStartLine + 1 ));
 						CursorWidth = ( CursorByteWidth - 1 +
 													MasterCursorWidth ) * 12;
 						CursorResized = 0;
@@ -727,7 +845,7 @@ DrawCursor()
 
 					if ( CursorViewable )
 						XFillRectangle ( dpy, TeletextScreen, CursorGC,
-							CursorX, CursorY, CursorWidth, CursorDepth );
+								CursorX, CursorY, CursorWidth, CursorDepth );
 				}
 			}
 			else
@@ -742,14 +860,14 @@ DrawCursor()
 				if ( CursorResized )
 				{
 					if ( CursorStartLine > 19 || ( CursorByteWidth == 0 &&
-						 MasterCursorWidth == 0 ) || CursorEndLine <
-														CursorStartLine )
+							MasterCursorWidth == 0 ) || CursorEndLine <
+															CursorStartLine )
 						CursorViewable = 0;
 					else
 						CursorViewable = 1;
 					CursorY = NewCursorY * 19 + CursorStartLine;
-					CursorDepth = MIN ( 19, CursorEndLine -
-													CursorStartLine + 1 );
+					CursorDepth = XBEEB_MIN ( 19, ( CursorEndLine -
+													CursorStartLine + 1 ));
 					CursorWidth = ( CursorByteWidth - 1 + MasterCursorWidth )
 																		* 12;
 					CursorResized = 0;
